@@ -7,7 +7,7 @@ import functools
 
 import matplotlib
 
-from skrt.core import Archive, Data
+import skrt.core
 import skrt.image
 import skrt.structures
 
@@ -266,6 +266,18 @@ class Dose(ImageOverlay):
             self.dose_type = getattr(ds, 'DoseType', None)
             self.dose_summation_type = getattr(ds, 'DoseSummationType', None)
 
+    def copy_dicom(self, *args, **kwargs):
+        """
+        Copy (single) source dicom file.
+
+        The present class (skrt.dose.Dose) inherits from skrt.image.Image,
+        which in turn inherits from skrt.core.Dated.  This method
+        forces calling of the copy_dicom() method of the latter.
+        For details of parameters, see documentation for:
+        skrt.core.Dated.copy_dicom().
+        """
+        skrt.core.Dated.copy_dicom(self, *args, **kwargs)
+
     def set_image(self, image):
         """Set associated image. Image.add_dose(self) will also be called."""
 
@@ -441,7 +453,7 @@ class Dose(ImageOverlay):
             return np.quantile(doses, quantile)
 
 
-class Plan(Archive):
+class Plan(skrt.core.Archive):
     def __init__(self, path="", load=True):
 
         self.loaded = False
@@ -459,11 +471,11 @@ class Plan(Archive):
         self.image = None
         self.structure_set = None
         self.doses = []
-        self.objectives = Data()
+        self.objectives = skrt.core.Data()
         for constraint_attribute in Constraint.get_weight_and_objectives():
             setattr(self.objectives, constraint_attribute, None)
 
-        Archive.__init__(self, path)
+        skrt.core.Archive.__init__(self, path)
 
         self.constraints_loaded = False
         if load:
@@ -550,21 +562,25 @@ class Plan(Archive):
                 continue
             roi.roi_type = item.DoseReferenceType
             roi.constraint = Constraint()
-            roi.constraint.weight = float(item.ConstraintWeight)
+            roi.constraint.weight = skrt.core.get_float(
+                    item, "ConstraintWeight")
             if 'ORGAN_AT_RISK' == item.DoseReferenceType:
-                roi.constraint.maximum_dose = float(item.OrganAtRiskMaximumDose)
-                roi.constraint.full_volume_dose = float(
-                        item.OrganAtRiskFullVolumeDose)
-                roi.constraint.overdose_volume_fraction = float(
-                        item.OrganAtRiskOverdoseVolumeFraction)
+                roi.constraint.maximum_dose = skrt.core.get_float(
+                        item, "OrganAtRiskMaximumDose")
+                roi.constraint.full_volume_dose = skrt.core.get_float(
+                        item, "OrganAtRiskFullVolumeDose")
+                roi.constraint.overdose_volume_fraction = skrt.core.get_float(
+                        item, "OrganAtRiskOverdoseVolumeFraction")
                 self.organs_at_risk.append(roi)
             elif 'TARGET' == item.DoseReferenceType:
-                roi.constraint.minimum_dose = float(item.TargetMinimumDose)
-                roi.constraint.prescription_dose = float(
-                        item.TargetPrescriptionDose)
-                roi.constraint.maximum_dose = float(item.TargetMaximumDose)
-                roi.constraint.underdose_volume_fraction = float(
-                        item.TargetUnderdoseVolumeFraction)
+                roi.constraint.minimum_dose = skrt.core.get_float(
+                        item, "TargetMinimumDose")
+                roi.constraint.prescription_dose = skrt.core.get_float(
+                        item, "TargetPrescriptionDose")
+                roi.constraint.maximum_dose = skrt.core.get_float(
+                        item, "TargetMaximumDose")
+                roi.constraint.underdose_volume_fraction = skrt.core.get_float(
+                        item, "TargetUnderdoseVolumeFraction")
                 self.targets.append(roi)
 
         self.constraints_loaded = True
@@ -741,7 +757,7 @@ class Plan(Archive):
         self.doses = []
 
 
-class Constraint(Data):
+class Constraint(skrt.core.Data):
     '''
     Container for data relating to a dose constraint.
     '''
