@@ -10,6 +10,7 @@ import timeit
 import pandas as pd
 
 import skrt.core
+from skrt.core import fullpath, get_data_indices, get_indexed_objs
 from skrt.image import Image
 from skrt.structures import StructureSet
 from skrt.dose import Dose, Plan, remove_duplicate_doses, sum_doses
@@ -493,15 +494,15 @@ class Study(skrt.core.Archive):
                     if plan_uid == referenced_plan.ReferencedSOPInstanceUID:
                         dose.set_plan(plan)
 
-    def save_images_as_nifti(self, out_dir='.', image_types=None, times=None,
-            verbose=True, image_size=None, voxel_size=None, fill_value=-1024,
+    def save_images_as_nifti(self, outdir='.', image_types=None, times=None,
+            verbose=True, image_size=None, voxel_size=None, fill_value=None,
             bands=None, require_structure_set=None):
         '''
         Save study's image data as nifti files.
 
         **Parameters:**
 
-        out_dir - str, default='.'
+        outdir - str, default='.'
             Directory to which nifti files will be saved.
 
         image_types - list/str/None, default=None
@@ -526,8 +527,9 @@ class Study(skrt.core.Archive):
             Voxel dimensions (dx, dy, dz) in mm for image resizing
             prior to writing.  If None, original voxel size is kept.
 
-        fill_value - int/float, default=-1024
-            Value used when extrapolating image outside data area>
+        fill_value - int/float, default=None
+            Value used when extrapolating image outside data area.
+            If None, use mininum value inside data area.
 
         bands - dict, default=None
             Nested dictionary of value bandings to be applied before
@@ -556,7 +558,7 @@ class Study(skrt.core.Archive):
             require_structure_set = []
 
         # Obtain full path to output directory.
-        out_dir = skrt.core.fullpath(out_dir)
+        outdir = skrt.core.fullpath(outdir)
 
         # Obtain set of image types to be saved.
         save_types = set(self.image_types)
@@ -566,9 +568,9 @@ class Study(skrt.core.Archive):
             save_types = save_types.intersection(set(image_types))
         if verbose:
             if save_types:
-                print(f"Saving imaging data to '{out_dir}':")
+                print(f"Saving image data to '{outdir}':")
             else:
-                print(f"No data to save for output directory '{out_dir}'")
+                print(f"No data to save for output directory '{outdir}'")
 
         # Loop over image types to be saved.
         for save_type in save_types:
@@ -598,8 +600,10 @@ class Study(skrt.core.Archive):
 
                 # Resize to required image size and voxel size.
                 if image_size is not None or voxel_size is not None:
+                    im_fill_value = (fill_value if fill_value is not None
+                            else out_image.get_min())
                     out_image.resize(image_size=image_size,
-                            voxel_size=voxel_size, fill_value=fill_value)
+                            voxel_size=voxel_size, fill_value=im_fill_value)
 
                 # Apply banding.
                 if save_type in bands:
@@ -607,12 +611,12 @@ class Study(skrt.core.Archive):
                         out_image.apply_bandings(bands[save_type])
 
                 # Define output path and write image.
-                outname = (f'{out_dir}/{save_type.upper()}'
+                outname = (f'{outdir}/{save_type.upper()}'
                            f'_{image.timestamp}_{idx+1:03}'
                            f'_{suffix}.nii.gz')
                 out_image.write(outname=outname, verbose=verbose)
 
-    def save_structure_sets_as_nifti(self, out_dir='.',
+    def save_structure_sets_as_nifti(self, outdir='.',
             image_types=None, times=None, files=None,
             verbose=True, image_size=None, voxel_size=None, roi_names=None,
             force_roi_nifti=False, bilateral_names=None):
@@ -622,7 +626,7 @@ class Study(skrt.core.Archive):
 
         **Parameters:**
 
-        out_dir - str, default='.'
+        outdir - str, default='.'
             Directory to which nifti files will be saved.
 
         image_types - list/str/None, default=None
@@ -671,7 +675,7 @@ class Study(skrt.core.Archive):
         '''
 
         # Obtain full path to output directory.
-        out_dir = skrt.core.fullpath(out_dir)
+        outdir = skrt.core.fullpath(outdir)
 
         # Initialise list or names for bilateral ROIs
         if bilateral_names is None:
@@ -685,9 +689,9 @@ class Study(skrt.core.Archive):
             save_types = save_types.intersection(set(image_types))
         if verbose:
             if save_types:
-                print(f"Saving structure_set data to '{out_dir}':")
+                print(f"Saving structure_set data to '{outdir}':")
             else:
-                print(f"No data to save for output directory '{out_dir}'")
+                print(f"No data to save for output directory '{outdir}'")
 
         # Loop over image types for which structure sets are to be saved.
         for save_type in save_types:
@@ -707,8 +711,8 @@ class Study(skrt.core.Archive):
                 # Only consider structure sets with requested indices.
                 if files and structure_sets:
                     if save_type in files and files[save_type]:
-                        structure_sets = [structure_sets[idx]
-                                for idx in files[save_type]]
+                        structure_sets = get_indexed_objs(
+                                structure_sets, files[save_type])
 
                 # Loop over structure sets
                 idx = 0
@@ -751,7 +755,7 @@ class Study(skrt.core.Archive):
                         if roi_name in ss.get_roi_names():
                             # Save ROI mask.
                             roi = ss.get_roi(roi_name)
-                            roi.write(outname=outname, outdir=out_dir,
+                            roi.write(outname=outname, outdir=outdir,
                                     ext='nii.gz', verbose=verbose)
                             idx_add = 1
                         elif force_roi_nifti:
@@ -760,11 +764,110 @@ class Study(skrt.core.Archive):
                             im = Image(ss.image)
                             shape = im.get_data().shape
                             im.data = np.zeros(shape)
-                            outname = f'{out_dir}/{outname}'
+                            outname = f'{outdir}/{outname}'
                             ss.image.write(outname=outname, verbose=verbose)
                             idx_add = 1
                     idx += idx_add
 
+    def copy_dicom(self, outdir=".",
+            images_to_copy=None, structure_sets_to_copy=None,
+            doses_to_copy=None, plans_to_copy=None, overwrite=True,
+            sort=True):
+        """
+        Copy study dicom data.
+
+        **Parameters:**
+
+        overwrite - bool, default=True
+            If False, skip images with pre-existing output directories.
+            If True, delete pre-existing output directories.
+
+        outdir - str, default='.'
+            Top-level directory to which nifti files for InnerEye will
+            be written for study.  Each output image will be in a separate
+            sub-directory, along with a file per associated ROI.
+
+        images_to_copy : list/str/dict, default=None
+            String specifiying image type for which all images are
+            to be copied; list of strings specifying image types
+            for which all images are to be copied; dictionary where
+            the keys are image types and the values are lists of
+            timestamp indices for the images to be copied, 0 being
+            the earliest and -1 being the most recent.  If set to
+            None, all images are copied.
+
+        structure_sets_to_copy : list/str/dict, default=None
+            String specifiying structure-set type for which all
+            structure sets are to be copied; list of strings
+            specifying structure-set types for which all structure
+            sets are to be copied; dictionary where the keys are
+            structure-set types and the values are lists of
+            timestamp indices for the structure sets to be copied,
+            0 being the earliest and -1 being the most recent.  If set to
+            None, all structure sets are copied.
+        """
+        # Ensure that study output directory exists.
+        study_dir = skrt.core.make_dir(outdir, overwrite=overwrite)
+
+        # Obtain dictionary associating indices to images.
+        image_indices = get_data_indices(images_to_copy, self.image_types)
+
+        # Identify non-image data.
+        non_image_data = [
+                ("RTSTRUCT", "structure_sets",
+                    get_data_indices(structure_sets_to_copy,
+                    getattr(self, "structure_set_types", []))),
+                ("RTDOSE", "doses", get_data_indices(doses_to_copy,
+                    getattr(self, "dose_types", []))),
+                ("RTPLAN", "plans", get_data_indices(plans_to_copy,
+                    getattr(self, "plan_types", []))),
+                ]
+
+        # Loop over image types.
+        # Overwriting taken into account at study level,
+        # so don't overwrite at sub-study level.
+        for image_type in image_indices:
+
+            # Loop over images of current type.
+            for idx1, im in enumerate(self.image_types[image_type]):
+
+                # Check that image can be loaded.
+                # If not, give warning and skip.
+                try:
+                    im.load()
+                except:
+                    im.data = None
+
+                if im.data is None:
+                    print(f"Problems loading: {image_type}")
+                    print(getattr(im, "dicom_paths", []))
+                    continue
+                
+                # Copy image.
+                im.copy_dicom_files(image_type, idx1, image_indices,
+                        study_dir / image_type.upper()
+                        / f"{im.timestamp}_{idx1+1:03}",
+                        overwrite=False, sort=sort)
+
+                # Copy non-image data that matches an image type.
+                for modality, reference, indices in non_image_data:
+                    if image_type in indices:
+                        for idx2, obj in enumerate(getattr(im, reference)):
+                            obj.copy_dicom_files(image_type, idx2, indices,
+                                    study_dir / modality / image_type.upper()
+                                    / f"{im.timestamp}_{idx1+1:03}",
+                                    overwrite=False, sort=sort)
+
+        # Copy non-image data not matching an image type.
+        for modality, reference, indices in non_image_data:
+            for data_type in indices:
+                if data_type not in image_indices:
+                    for idx3, obj in enumerate(
+                            getattr(self, f"{data_type}_{reference}")):
+                        obj.copy_dicom_files(data_type, idx3, indices,
+                                study_dir / modality / data_type.upper(),
+                                overwrite=False, sort=sort)
+
 
 class Patient(skrt.core.PathData):
     """
@@ -887,6 +990,9 @@ class Patient(skrt.core.PathData):
                 "structure_set": StructureSet,
                 }
 
+        # Modalities not currently handled.
+        unhandled_modalities = ["pr", "rtrecord", "sr"]
+
         # Loop through files in the directory pointed to path attribute,
         # and in sub-directories.  Use each file to try to instantiate
         # a skrt.core.DicomFile object.  Use instantiated objects to
@@ -897,7 +1003,8 @@ class Patient(skrt.core.PathData):
         file_paths = sorted(patient_path.glob("**/*"))
         for file_path in file_paths:
             dcm = skrt.core.DicomFile(file_path)
-            if dcm.ds:
+            if (dcm.ds and (dcm.study_instance_uid is not None)
+                and (dcm.modality not in unhandled_modalities)):
                 # Add to patient's list of Study objects.
                 matched_attributes = dcm.get_matched_attributes(self.studies,
                         "study_instance_uid")
@@ -2041,6 +2148,60 @@ class Patient(skrt.core.PathData):
             elif 'RTSTRUCT' == modality:
                 item.write(outdir=item_dir, ext=ext)
 
+
+    def copy_dicom(self, outdir=".", studies_to_copy=None, overwrite=True,
+            **kwargs):
+        """
+        Copy patient dicom data.
+
+        **Parameters:**
+        outdir - str, default='.'
+            Top-level output directory.  Within this, there will be a
+            patient sub-directory, containing a sub-directory for each
+            study, containing in turn a sub-directory for each data
+            modality.
+
+        studies_to_copy: list/dict, default=None
+            List of indices of studies for which data are
+            to be written, 0 being the earliest study and -1 being
+            the most recent: or a dictionary where keys will be used
+            as names of subdirectories grouping studies, and values
+            are indices of studies to be grouped.  If set to
+            None, data for all studies are written, keeping any
+            existing grouping.
+
+        overwrite : bool, default=True
+            If True, delete and recreate patient sub_directory
+            before copying files.  If False and the patient sub-directory
+            exists already, copy files to the existing directory.
+
+        **kwargs
+            Keyword arguments passed on to
+                skrt.patient.Study.copy_dicom().
+            For details, see this method's documentation.
+        """
+        # Define patient output directory.
+        patient_dir = skrt.core.make_dir(Path(fullpath(outdir)) / self.id,
+                overwrite=overwrite)
+
+        # If studies_to_copy is None, set to select all studies.
+        if studies_to_copy is None:
+            studies_to_copy = {}
+            for idx, study in enumerate(self.studies):
+                if not study.subdir in studies_to_copy:
+                    studies_to_copy[study.subdir] = []
+                studies_to_copy[study.subdir].append(idx)
+        elif not isinstance(studies_to_copy, dict):
+            studies_to_copy = {"": studies_to_copy}
+
+        # Process selected studies.
+        for group, indices in studies_to_copy.items():
+            for study in get_indexed_objs(self.studies, indices):
+                # Overwriting taken into account at patient level,
+                # so don't overwrite at study level.
+                study.copy_dicom(outdir=patient_dir / group / study.timestamp,
+                        overwrite=False, **kwargs)
+
 def find_matching_object(obj, possible_matches):
     """For a given object <obj> and a list of potential matching objects
     <possible_matches>, find an object that matches <obj>'s path,
